import socket


def has_internet(host="duckduckgo.com", port=53, timeout=3):
    """Check for internet connectivity by attempting to connect to a reliable host."""
    try:
        socket.setdefaulttimeout(timeout)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.connect((host, port))
        return True
    except OSError:
        return False



import pytest
from astropy.coordinates import SkyCoord

from cabaret.queries import GaiaQuery

from .utils import has_internet


@pytest.mark.skipif(not has_internet(), reason="Requires internet")
def test_get_sources_basic():
    center = SkyCoord(ra=10.68458, dec=41.26917, unit="deg")
    radius = 0.05
    sources = GaiaQuery.get_sources(center, radius, limit=10, timeout=30)
    assert len(sources) <= 10
    assert sources is not None


@pytest.mark.skipif(not has_internet(), reason="Requires internet")
def test_get_sources_timeout():
    center = SkyCoord(ra=10.68458, dec=41.26917, unit="deg")
    radius = 0.05
    with pytest.raises(TimeoutError):
        GaiaQuery.get_sources(center, radius, limit=10, timeout=0.0001)



from dataclasses import dataclass


@dataclass
class Site:
    """
    Observatory site configuration.
    """

    sky_background: float = 150
    """Sky background in e-/m^2/arcsec^2/s."""

    seeing: float = 1.3
    """Atmospheric seeing in arcseconds."""

    latitude: float | None = None
    """Site latitude in degrees."""

    longitude: float | None = None
    """Site longitude in degrees."""

    elevation: float | None = None
    """Site elevation in meters."""

    sun_altitude: float | None = None
    """Sun altitude in degrees."""



import doctest

import matplotlib
import pytest

matplotlib.use("Agg")

import cabaret.camera
import cabaret.focuser
import cabaret.observatory
import cabaret.queries
import cabaret.site
import cabaret.sources
import cabaret.telescope

from .utils import has_internet

pytestmark = pytest.mark.filterwarnings("ignore:.*FigureCanvasAgg.*:UserWarning")


@pytest.mark.parametrize(
    "mod",
    [
        pytest.param(cabaret.sources, id=cabaret.sources.__name__),
        pytest.param(
            cabaret.queries,
            marks=pytest.mark.skipif(not has_internet(), reason="Requires internet"),
            id=cabaret.queries.__name__,
        ),
        pytest.param(cabaret.camera, id=cabaret.camera.__name__),
        pytest.param(cabaret.site, id=cabaret.site.__name__),
        pytest.param(cabaret.focuser, id=cabaret.focuser.__name__),
        pytest.param(cabaret.telescope, id=cabaret.telescope.__name__),
        pytest.param(
            cabaret.observatory,
            marks=pytest.mark.skipif(not has_internet(), reason="Requires internet"),
            id=cabaret.observatory.__name__,
        ),
    ],
)
def test_doctests(mod):
    flags = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    result = doctest.testmod(mod, optionflags=flags)
    assert result.failed == 0, f"{result.failed} doctest failures in {mod.__name__}"



import importlib.util

if not importlib.util.find_spec("matplotlib"):
    raise ImportError("Please install matplotlib to use cabaret.utils.")


import matplotlib.pyplot as plt
from astropy.visualization import ZScaleInterval


def plot_image(
    image,
    title: str | None = None,
    cmap="gray",
    ax=None,
    add_colorbar=True,
    colorbar_kwargs={},
    transparent=True,
    contrast=0.25,
):
    """Plot a 2D image with zscale normalization.

    Parameters
    ----------
    image : np.ndarray
        The image to plot.
    title : str, optional
        Title for the plot.
    cmap : str, optional
        Colormap to use.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on.
    add_colorbar : bool, optional
        Whether to add a colorbar.
    colorbar_kwargs : dict, optional
        Additional kwargs for colorbar.
    transparent : bool, optional
        If True, set figure background to transparent.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 8))
    else:
        fig = plt.gcf()
    interval = ZScaleInterval(contrast=contrast)
    vmin, vmax = interval.get_limits(image)
    img = ax.imshow(image, vmin=vmin, vmax=vmax, cmap=cmap, rasterized=True)
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel("Pixels")
    ax.set_ylabel("Pixels")

    if add_colorbar:
        cbar = plt.colorbar(
            img, ax=ax, **{"fraction": 0.046, "pad": 0.04} | colorbar_kwargs
        )
        cbar.set_label("ADU")

    if transparent:
        fig.patch.set_alpha(0)
        ax.patch.set_alpha(0)

    return img



from concurrent.futures import ThreadPoolExecutor, TimeoutError
from datetime import datetime
from enum import Enum

import numpy as np
from astropy.coordinates import Angle, SkyCoord
from astropy.table import Table
from astropy.time import Time
from astropy.units import Quantity

from cabaret.sources import Sources

__all__ = [
    "Filters",
    "GaiaQuery",
]


class Filters(Enum):
    """Allowed Gaia and 2MASS flux filter_band strings.

    Examples
    --------
    >>> from cabaret.queries import Filters
    >>> Filters.G
    <Filters.G: 'phot_g_mean_flux'>
    >>> Filters.from_string('RP')
    <Filters.RP: 'phot_rp_mean_flux'>
    >>> Filters.is_tmass('J')
    True
    >>> Filters.options()
    ('G', 'BP', 'RP', 'J', 'H', 'KS')
    """

    G = "phot_g_mean_flux"
    """ Gaia G band flux in [e-/s] """
    BP = "phot_bp_mean_flux"
    """ Gaia BP band flux in [e-/s] """
    RP = "phot_rp_mean_flux"
    """ Gaia RP band flux in [e-/s] """
    J = "j_m"
    """ 2MASS J band magnitude """
    H = "h_m"
    """ 2MASS H band magnitude """
    KS = "ks_m"
    """ 2MASS KS band magnitude """

    @classmethod
    def options(cls) -> tuple[str, ...]:
        """Return all valid filter_band options."""
        return tuple(cls.__members__.keys())

    @classmethod
    def from_string(cls, value: str) -> "Filters":
        """Return the Filters enum member for a given string."""
        try:
            return cls[value.upper()]
        except KeyError:
            raise ValueError(
                f"Invalid filter_band string: {value}. "
                f"Valid options are: {cls.options()}"
            )

    @classmethod
    def is_tmass(cls, value: "Filters | str") -> bool:
        """Check if the filter_band string is a 2MASS filter_band."""
        if isinstance(value, cls):
            return value.name in ("J", "H", "KS")
        elif isinstance(value, str):
            return value.upper() in ("J", "H", "KS")
        else:
            raise ValueError(
                f"Value must be an Filters enum or string, got {type(value)}"
            )

    @classmethod
    def ensure_enum(cls, value: "Filters | str") -> "Filters":
        """Convert a string or Filters to Filters enum."""
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            return cls.from_string(value)
        else:
            raise ValueError(
                f"Value must be a Filters enum or string, got {type(value)}"
            )

    @classmethod
    def is_valid(cls, value: str) -> bool:
        """Check if the filter_band string is valid."""
        return value.upper() in cls.__members__


class GaiaQuery:
    """Class to query the Gaia archive and retrieve sources.

    The class provides methods to query Gaia and return either the raw Astropy Table
    or a Sources instance with RA-DEC coordinates and fluxes.

    Examples
    --------
    >>> from cabaret.queries import GaiaQuery
    >>> from astropy.coordinates import SkyCoord
    >>> center = SkyCoord(ra=10.68458, dec=41.269, unit="deg")

    The Astropy Table from Gaia can be obtained with:

    >>> table = GaiaQuery.query(center, radius=0.05, limit=10, timeout=30)

    Whereas a Sources instance carrying coordinates and fluxes can be queried for with:

    >>> sources = GaiaQuery.get_sources(center, radius=0.05, limit=10, timeout=30)
    """

    @staticmethod
    def query(
        center: tuple[float, float] | SkyCoord,
        radius: float | Angle,
        filter_band: Filters = Filters.G,
        limit: int = 100000,
        timeout: float | None = None,
    ) -> Table:
        """Query Gaia Archive within a given radius around the center position.

        Parameters
        ----------
        center : tuple or astropy.coordinates.SkyCoord
            The sky coordinates of the center of the FOV.
            If a tuple is given, it should contain the RA and DEC in degrees.
        radius : float or astropy.units.Quantity
            The radius of the FOV in degrees. If a Quantity is given, it must be
            convertible to degrees.¨
        filter_band : Filters or str, optional
            The filter to use for the flux column. Default is Filters.G.
        limit : int, optional
            The maximum number of sources to retrieve from the Gaia archive.
            By default, it is set to 100000.
        timeout : float, optional
            The maximum time to wait for the Gaia query to complete, in seconds.
            If None, there is no timeout. By default, it is set to None.

        Returns
        -------
        astropy.table.Table
            The raw Astropy Table returned by the Gaia archive.

        Example
        -------
        >>> from cabaret.queries import GaiaQuery
        >>> from astropy.coordinates import SkyCoord
        >>> center = SkyCoord(ra=10.68458, dec=41.26917, unit='deg')
        >>> table = GaiaQuery.query(center, radius=0.1, limit=10, timeout=30)
        """
        filter_band = Filters.ensure_enum(filter_band)

        if isinstance(center, SkyCoord):
            ra = center.ra.deg  # type: ignore
            dec = center.dec.deg  # type: ignore
        else:
            ra, dec = center

        gaia_flux_column = (
            filter_band.value
            if not Filters.is_tmass(filter_band.name)
            else Filters.G.value
        )
        select_cols = [
            "gaia.ra",
            "gaia.dec",
            "gaia.pmra",
            "gaia.pmdec",
            f"{gaia_flux_column}",
        ]
        joins = []
        where = []
        order_by = f"{filter_band.value} DESC"  # Prefer brighter stars

        if Filters.is_tmass(filter_band.name):
            select_cols.append(filter_band.value)
            joins.extend(
                [
                    "INNER JOIN gaiadr2.tmass_best_neighbour AS tmass_match "
                    + "ON tmass_match.source_id = gaia.source_id",
                    "INNER JOIN gaiadr1.tmass_original_valid AS tmass "
                    + "ON tmass.tmass_oid = tmass_match.tmass_oid",
                ]
            )
            order_by = f"{filter_band.value} ASC"
            where.append(f"{filter_band.value} IS NOT NULL")
        else:
            where.append(f"{filter_band.value} IS NOT NULL")

        radius = radius.value if isinstance(radius, Quantity) else float(radius)
        where.append(
            f"1=CONTAINS(POINT('ICRS', {ra}, {dec}), "
            f"CIRCLE('ICRS', gaia.ra, gaia.dec, {radius}))"
        )

        select_clause = ", ".join(select_cols)
        joins_clause = "\n".join(joins)
        where_clause = " AND ".join(where)

        query = f"""
        SELECT TOP {limit} {select_clause}
        FROM gaiadr2.gaia_source AS gaia
        {joins_clause}
        WHERE {where_clause}
        ORDER BY {order_by}
        """

        table = GaiaQuery._launch_job_with_timeout(query, timeout=timeout)
        return table

    @staticmethod
    def get_sources(
        center: tuple[float, float] | SkyCoord,
        radius: float | Angle,
        filter_band: Filters | str = Filters.G,
        dateobs: datetime | None = None,
        limit: int = 100000,
        timeout: float | None = None,
    ) -> Sources:
        """
        Query the Gaia archive to retrieve the RA-DEC coordinates of stars
        within a given radius of a center position, along with their fluxes.

        Parameters
        ----------
        center : tuple or astropy.coordinates.SkyCoord
            The sky coordinates of the center of the FOV.
            If a tuple is given, it should contain the RA and DEC in degrees.
        radius : float or astropy.units.Quantity
            The radius of the FOV in degrees. If a Quantity is given, it must be
            convertible to degrees.
        filter_band : Filters or str, optional
            The filter to use for the flux column. Default is Filters.G.
        dateobs : datetime.datetime, optional
            The date of the observation. If given, the proper motions of the sources
            will be taken into account. By default, it is set to None.
        limit : int, optional
            The maximum number of sources to retrieve from the Gaia archive.
            By default, it is set to 10000.
        timeout : float, optional
            The maximum time to wait for the Gaia query to complete, in seconds.
            If None, there is no timeout. By default, it is set to None.

        Returns
        -------
        Sources
            A Sources instance containing the coordinates and fluxes of the retrieved
            sources.


        Notes
        -----
        If `filter_band` is a 2MASS filter (J, H, KS), the fluxes are calculated
        from the 2MASS magnitudes using `cabaret.queries.tmass_mag_to_photons`.

        Raises
        ------
        ImportError
            If the astroquery package is not installed.

        Examples
        --------
        >>> from cabaret.queries import GaiaQuery
        >>> from astropy.coordinates import SkyCoord
        >>> center = SkyCoord(ra=10.68458, dec=41.26917, unit='deg')
        >>> sources = GaiaQuery.get_sources(
        ...     center, radius=0.1, timeout=30
        ... )  # doctest: +SKIP
        Sources(coords=<SkyCoord (ICRS): (ra, dec) in deg
            [(10.63950247, 41.26393165), (10.6880729 , 41.22524785),
            (10.70349581, 41.25357386), (10.70022208, 41.26019689),
            (10.71333998, 41.29943347), (10.73974676, 41.2942209 ),
            (10.71181048, 41.29130279), (10.68780207, 41.31717482),
            (10.63804045, 41.27468757), (10.64397532, 41.25237352)]>, fluxes=array(
                [169435.62814443,  52203.9396396 ,  41716.18126449,  29035.89106422,
                22990.85994301,  17672.53437883,  15953.21022642,  15077.12262318,
                14004.42013396,  12271.11779953]))

        """
        filter_band = Filters.ensure_enum(filter_band)

        table = GaiaQuery.query(
            center=center,
            radius=radius,
            limit=limit,
            timeout=timeout,
            filter_band=filter_band,
        )

        if dateobs is not None:
            table = GaiaQuery._apply_proper_motion(table, dateobs)

        if Filters.is_tmass(filter_band.name):
            fluxes = GaiaQuery._tmass_mag_to_photons(
                table[filter_band.value].value.data,  # type: ignore
                filter_band,
            )
        else:
            fluxes = table[filter_band.value].value.data  # type: ignore
        table.remove_rows(np.isnan(fluxes))
        fluxes = fluxes[~np.isnan(fluxes)]

        return Sources.from_arrays(
            ra=table["ra"].value.data,  # type: ignore
            dec=table["dec"].value.data,  # type: ignore
            fluxes=fluxes,
        )

    @staticmethod
    def _launch_job_with_timeout(query, timeout=None, **kwargs) -> Table:
        """
        Launch a Gaia job and return its results, optionally enforcing a timeout.

        Parameters
        ----------
        query : str
            The query string passed to Gaia.launch_job.
        timeout : float or None, optional
            Maximum number of seconds to wait for Gaia.launch_job to complete.
            If None, the job is run on the main thread (no thread overhead).
        **kwargs
            Additional keyword arguments forwarded to Gaia.launch_job.

        Returns
        -------
        object
            The result returned by job.get_results().

        Raises
        ------
        TimeoutError
            If `timeout` is not None and the call does not complete within `timeout`.
        """
        from astroquery.gaia import Gaia

        # Run directly on the main thread when no timeout is requested to avoid
        # unnecessary thread creation and to preserve original callstacks/tracebacks.
        if timeout is None:
            job = Gaia.launch_job(query, **kwargs)
            return job.get_results()  # type: ignore

        with ThreadPoolExecutor() as executor:
            future = executor.submit(Gaia.launch_job, query, **kwargs)
            try:
                job = future.result(timeout=timeout)
                return job.get_results()  # type: ignore
            except TimeoutError:
                raise TimeoutError(
                    "Gaia query timed out."
                    " You may want to increase the timeout or reduce the query size."
                    f"Query was: {query}"
                )

    @staticmethod
    def _tmass_mag_to_photons(mags: np.ndarray, filter_band: Filters) -> np.ndarray:
        """Convert 2MASS J magnitudes to photon fluxes at mag 0.

        Reference: https://lweb.cfa.harvard.edu/~dfabricant/huchra/ay145/mags.html
        Returns photons/sec/m^2 for each magnitude.
        """
        # Use a dict for 2MASS filter properties
        dlam_lam_map = {"J": 0.16, "H": 0.23, "KS": 0.23}
        flux_m0_map = {"J": 1600, "H": 1080, "KS": 670}

        try:
            dlam_lam = dlam_lam_map[filter_band.name]
            flux_m0 = flux_m0_map[filter_band.name]
        except KeyError:
            raise ValueError(
                f"tmass_mag_to_photons expects a 2MASS filter (J,H,KS), "
                f"got {filter_band}."
            )

        Jy = 1.51e7  # [photons sec^-1 m^-2 (dlambda/lambda)^-1]
        photons = dlam_lam * flux_m0 * Jy  # [photons sec^-1 m^-2] at mag 0
        return photons * 10 ** (-0.4 * mags)

    @staticmethod
    def _apply_proper_motion(table: Table, dateobs: datetime | Time) -> Table:
        """
        Apply proper motion correction to RA and DEC columns
        for the given observation date.
        """

        if isinstance(dateobs, Time):
            dateobs_frac = float(dateobs.decimalyear)
        elif isinstance(dateobs, datetime):
            dateobs_frac = dateobs.year + (dateobs.timetuple().tm_yday - 1) / 365.25  # type: ignore
        else:
            raise ValueError(
                f"dateobs must be an astropy.time.Time or datetime, got {type(dateobs)}"
            )

        years = dateobs_frac - 2015.5  # type: ignore
        table["ra"] += years * table["pmra"] / 1000 / 3600  # type: ignore
        table["dec"] += years * table["pmdec"] / 1000 / 3600  # type: ignore

        return table



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.random
from astropy.coordinates import Angle
from astropy.wcs import WCS


@dataclass
class Camera:
    """
    Camera configuration and properties.

    Examples
    --------
    Create a camera with multiple pixel defects:

    >>> from cabaret import Camera, Observatory, Sources
    >>> pixel_defects = {
    ...     "cold": {"type": "constant", "value": 0, "rate": 0.01, "seed": 1},
    ...     "noise": {"type": "noise", "rate": 0.02, "noise_level": 100, "seed": 2},
    ...     "qe": {"type": "quantum_efficiency_map", "seed": 3},
    ...     "smear": {"type": "readout_smear", "readout_time": 1},
    ... }
    >>> camera = Camera(width=1024, height=1024, pixel_defects=pixel_defects)
    >>> observatory = Observatory(camera=camera)
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center

    Now generate an image with and without pixel defects:

    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     ra=ra, dec=dec, exp_time=10, seed=42, sources=sources,
    ...     convert_all_to_adu=True
    ... )

    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 2, figsize=(7, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with multiple defects")
    >>> plt.subplots_adjust(wspace=0.1)
    >>> plt.show()
    """

    name: str
    """Name of the camera."""

    width: int
    """Number of pixels in the x-direction (width)."""

    height: int
    """Number of pixels in the y-direction (height)."""

    bin_x: int
    """Binning factor in x."""

    bin_y: int
    """Binning factor in y."""

    pitch: float
    """Pixel pitch in microns."""

    plate_scale: float | None
    """Arcseconds per pixel (if None, calculated from pitch and telescope)."""

    max_adu: int
    """Maximum ADU value."""

    well_depth: int
    """Full well capacity in electrons."""

    bias: int
    """Bias level in ADU."""

    gain: float
    """Gain in electrons per ADU."""

    read_noise: float
    """Read noise in electrons."""

    dark_current: float
    """Dark current in electrons per second."""

    average_quantum_efficiency: float
    """Average quantum efficiency (fraction)."""

    rotation: float
    """Rotation angle of the camera in degrees."""

    pixel_defects: dict
    """Dictionary of pixel defect configurations."""

    def __init__(
        self,
        name: str = "gaia-camera-simulated",
        width: int = 1024,
        height: int = 1024,
        bin_x: int = 1,
        bin_y: int = 1,
        pitch: float = 13.5,
        plate_scale: float | None = None,
        max_adu: int = 2**16 - 1,
        well_depth: int = 2**16 - 1,
        bias: int = 300,
        gain: float = 1.0,
        read_noise: float = 6.2,
        dark_current: float = 0.2,
        average_quantum_efficiency: float = 0.8,
        rotation: float = 0.0,
        pixel_defects: dict | None = None,
    ) -> None:
        self.name = name
        self.width = width
        self.height = height
        self.bin_x = bin_x
        self.bin_y = bin_y
        self.pitch = pitch
        self.plate_scale = plate_scale
        self.max_adu = max_adu
        self.well_depth = well_depth
        self.bias = bias
        self.gain = gain
        self.read_noise = read_noise
        self.dark_current = dark_current
        self.average_quantum_efficiency = average_quantum_efficiency
        self.rotation = rotation

        if pixel_defects:
            self.pixel_defects = {
                key: (
                    self._create_pixel_defect(key, **value)
                    if isinstance(value, dict)
                    else value
                )
                for key, value in pixel_defects.items()
            }
        else:
            self.pixel_defects = {}

    @property
    def shape(self) -> tuple[int, int]:
        """Tuple of (height, width) of the camera in pixels."""
        return (self.height, self.width)

    @property
    def size(self) -> int:
        """Total number of pixels in the camera."""
        return self.height * self.width

    def make_base_image(
        self, exp_time, rng: np.random.Generator | None = None
    ) -> np.ndarray:
        """Generate the detector base image (bias, dark, read noise).

        Parameters
        ----------
        exp_time : float
            Exposure time in seconds.
        rng : np.random.Generator, optional
            Random number generator for reproducibility. If None, a new generator
            will be created.

        Returns
        -------
        np.ndarray
            The base image with bias, dark current, and read noise added.
        """
        if rng is None:
            rng = np.random.default_rng()

        base = np.ones((self.height, self.width)).astype(np.float64)

        base += rng.poisson(base * self.dark_current * exp_time).astype(np.float64)

        base += rng.normal(0, self.read_noise, (self.height, self.width)).astype(
            np.float64
        )

        return base

    def apply_pixel_defects(self, image: np.ndarray, exp_time: float) -> np.ndarray:
        """Apply pixel defects to the image.

        Parameters
        ----------
        image : np.ndarray
            The image to which the pixel defects will be applied.
        exp_time : float
            Exposure time in seconds.

        Returns
        -------
        np.ndarray
            The image with pixel defects applied.
        """
        for defect in self.pixel_defects.values():
            image = defect.introduce_pixel_defect(
                image=image, camera=self, exp_time=exp_time
            )
        return image

    def to_adu_image(self, image: np.ndarray) -> np.ndarray:
        """Convert to ADU, add bias, clip, and cast to uint16.

        Parameters
        ----------
        image : np.ndarray
            The image to be finalized.

        Returns
        -------
        np.ndarray
            The finalized image as a uint16 numpy array.

        """
        image = image / self.gain + self.bias

        image = np.clip(image, 0, self.max_adu)

        return image.astype(np.uint16)

    def bin_image(self, image: np.ndarray) -> np.ndarray:
        """Bin the image according to the camera's binning factors.

        Parameters
        ----------
        image : np.ndarray
            The image to be binned.

        Returns
        -------
        np.ndarray
            The binned image.
        """
        if self.bin_x == 1 and self.bin_y == 1:
            return image

        binned_height = image.shape[0] // self.bin_y
        binned_width = image.shape[1] // self.bin_x

        binned_image = (
            image[: binned_height * self.bin_y, : binned_width * self.bin_x]
            .reshape(binned_height, self.bin_y, binned_width, self.bin_x)
            .sum(axis=(1, 3))
        )

        return binned_image

    def set_plate_scale_from_focal_length(self, focal_length: float):
        """Set the plate scale based on the focal length of the telescope.

        Parameters
        ----------
        focal_length : float
            Focal length of the telescope in m.

        Returns
        -------
        None
        """
        self.plate_scale = (
            2
            * np.arctan((self.pitch * 1e-6) / (2 * focal_length))
            * (180 / np.pi)
            * 3600
        )  # "/pixel

    def _create_pixel_defect(
        self,
        name: str,
        type: Literal["constant", "column", "noise", "fixed_pattern"] = "constant",
        **kwargs,
    ) -> "PixelDefect":
        defect_classes = {
            "constant": ConstantPixelDefect,
            "column": ColumnPixelDefect,
            "noise": RandomNoisePixelDefect,
            "quantum_efficiency_map": QuantumEfficiencyMapPixelDefect,
            "readout_smear": ReadoutSmearPixelDefect,
        }
        defect_type = type

        if defect_type not in defect_classes:
            raise ValueError(f"Unknown pixel defect type for {name}")

        return defect_classes[defect_type](name=name, **kwargs)

    @classmethod
    def create_ideal_camera(cls, **kwargs) -> "Camera":
        """Create a defect-free camera."""
        parameters = {
            "read_noise": 0,
            "dark_current": 0,
            "average_quantum_efficiency": 1.0,
            "bias": 0,
            "gain": 1.0,
        }
        return cls(**(parameters | kwargs))

    def get_wcs(self, center):
        """Get a WCS object for the camera centered at the given sky coordinate."""
        if self.plate_scale is None:
            raise ValueError("plate_scale must be set to compute WCS.")

        # Convert plate scale to degrees per pixel
        scale_deg = self.plate_scale / 3600  # arcsec to deg

        theta = np.deg2rad(self.rotation)
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        # CD matrix with rotation
        cd11 = -scale_deg * cos_theta
        cd12 = scale_deg * sin_theta
        cd21 = -scale_deg * sin_theta
        cd22 = -scale_deg * cos_theta

        wcs = WCS(naxis=2)
        wcs.wcs.cd = np.array([[cd11, cd12], [cd21, cd22]])
        wcs.wcs.cunit = ["deg", "deg"]
        wcs.wcs.crpix = [int(self.width / 2), int(self.height / 2)]
        wcs.wcs.crval = [center.ra.deg, center.dec.deg]
        wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        return wcs

    def get_fov_radius(self) -> Angle:
        """Half-diagonal field radius in degrees."""
        if self.plate_scale is None:
            raise ValueError("plate_scale must be set to compute FOV.")
        fov_y = self.height * self.plate_scale / 3600.0
        fov_x = self.width * self.plate_scale / 3600.0
        return Angle(np.sqrt(fov_x**2 + fov_y**2) / 2, "deg")

    def _create_blank_image(self) -> np.ndarray:
        """Create a blank image filled with zeros for testing."""
        return np.zeros((self.height, self.width), dtype=np.uint16)

    def on_camera_mask(self, pixel_coords: np.ndarray) -> np.ndarray:
        """Clip the given pixel coordinates to the camera's bounds.

        Parameters
        ----------
        pixel_coords : np.ndarray
            An array of shape (n, 2) containing the pixel coordinates to be clipped.

        Returns
        -------
        np.ndarray
            An array of shape (m, 2) containing the pixel coordinates within the
            camera's bounds.
        """
        in_camera = (
            (pixel_coords[0, :] >= 0)
            & (pixel_coords[0, :] < self.width)
            & (pixel_coords[1, :] >= 0)
            & (pixel_coords[1, :] < self.height)
        )
        return in_camera


@dataclass
class PixelDefect(ABC):
    name: str = "defect"
    """Name of the defect."""

    rate: float = 0.0
    """Fraction of pixels affected by the defect."""

    seed: int = 0
    """Random seed for reproducibility."""

    _rng: numpy.random.Generator | None = None
    """Random number generator instance."""

    _pixels: np.ndarray | None = None
    """Array of pixel coordinates affected by the defect."""

    @property
    def pixels(self) -> np.ndarray:
        """Pixel coordinates affected by the defect."""
        if self._pixels is None:
            raise ValueError(f"{self.name} pixels are not defined.")
        return self._pixels

    @property
    def rng(self) -> numpy.random.Generator:
        """Random number generator instance."""
        if self._rng is None:
            self._rng = numpy.random.default_rng(self.seed)

        return self._rng

    @abstractmethod
    def introduce_pixel_defect(
        self, image: np.ndarray, camera: Camera, **kwargs
    ) -> np.ndarray:
        """Introduce the defect into the image.

        Parameters
        ----------
        image : np.ndarray
            The image to which the defect will be introduced.
        camera : Camera
            The camera to which the defect applies.
        """
        raise NotImplementedError

    def set_pixels(self, pixels: np.ndarray, camera: Camera):
        """Set the pixels for the defect.

        Parameters
        ----------
        pixels : np.ndarray
            The pixel coordinates of the defect.
        camera : Camera
            The camera to which the defect applies.
        """
        self._check_pixel_bounds(pixels, camera.height, camera.width, self.name)
        self._pixels = pixels

    def number_of_defect_pixels(self, camera: Camera) -> int:
        """Calculate the number of pixels affected by the defect."""
        return int(round(self.rate * camera.width * camera.height))

    def _select_random_pixels(self, camera: Camera) -> np.ndarray:
        return self.rng.integers(
            [camera.height, camera.width],
            size=(self.number_of_defect_pixels(camera), 2),
        )

    @staticmethod
    def _overwrite_pixel_values(
        image: np.ndarray, pixels: np.ndarray, pixel_values: int | np.ndarray
    ):
        if isinstance(pixel_values, np.ndarray):
            if not pixel_values.size == pixels.shape[0]:
                raise ValueError("Pixel values must match the number of pixels.")

        image[pixels[:, 0], pixels[:, 1]] = pixel_values

    @staticmethod
    def _check_pixel_bounds(pixels, height: int, width: int, name: str):
        if pixels is None:
            raise ValueError(f"{name} pixels are not defined.")
        if isinstance(pixels, list):
            pixels = np.array(pixels)
        if not pixels.ndim == 2 or not pixels.shape[1] == 2:
            raise ValueError(
                f"{name} pixels must be a numpy array of (x, y) tuples."
                f"Got shape {pixels.shape} instead."
            )
        if np.any(pixels[:, 0] >= height) or np.any(pixels[:, 1] >= width):
            raise ValueError(f"{name} pixels are outside the frame.")


@dataclass
class ConstantPixelDefect(PixelDefect):
    """A pixel defect that sets selected pixels to a constant value.

    Examples
    --------
    >>> from cabaret import Observatory, Sources
    >>> pixel_defects = {
    ...     "hot": {"type": "constant", "value": 5000, "rate": 0.01, "seed": 0},
    ...     "cold": {"type": "constant", "value": 0, "rate": 0.01, "seed": 1}
    ... }
    >>> observatory = Observatory(camera={"pixel_defects": pixel_defects})
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center
    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     exp_time=3, ra=ra, dec=dec, sources=sources, convert_all_to_adu=True
    ... )

    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 2, figsize=(7, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with constant defects")
    >>> plt.subplots_adjust(wspace=0.1)
    >>> plt.show()
    """

    value: int = 0
    """Constant value to assign to defect pixels."""

    seed: int = 0
    """Random seed for reproducibility."""

    def introduce_pixel_defect(self, image, camera, **kwargs):
        if self._pixels is None:
            self.set_pixels(self._select_random_pixels(camera), camera)

        self._overwrite_pixel_values(image, self.pixels, self.value)

        return image


@dataclass
class ColumnPixelDefect(PixelDefect):
    """A pixel defect that simulates column noise by selecting entire rows or
    columns of pixels to be defective.

    Examples
    --------
    >>> from cabaret import Observatory, Sources
    >>> pixel_defects = {
    ...     "column": {"type": "column", "value": 20, "rate": 0.01, "dim": 1}
    ... }
    >>> observatory = Observatory(camera={"pixel_defects": pixel_defects})
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center
    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     exp_time=3, ra=ra, dec=dec, sources=sources, convert_all_to_adu=True
    ... )


    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 2, figsize=(7, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with column defects")
    >>> plt.subplots_adjust(wspace=0.1)
    >>> plt.show()
    """

    value: int = 0
    """Constant value to assign to defect pixels."""

    seed: int = 0
    """Random seed for reproducibility."""

    dim: int = 0
    """Dimension along which to apply the defect."""

    _lines: np.ndarray | None = None

    @property
    def lines(self) -> np.ndarray:
        """Defective lines (rows or columns)."""
        if self._lines is None:
            raise ValueError(f"{self.name} lines are not defined.")
        return self._lines

    def introduce_pixel_defect(self, image, camera, **kwargs):
        if self._lines is None:
            self.set_lines(self._select_random_lines(camera, self.dim), camera)

        self._overwrite_pixel_values(image, self.pixels, self.value)

        return image

    def set_lines(self, lines: np.ndarray | list, camera: Camera):
        """Set the lines for the defect."""
        self._lines = np.array(lines)

        # Set pixels based on the selected lines
        if self.dim == 0:
            if not np.all(self._lines < camera.height):
                raise ValueError("Selected lines are outside the frame.")
            X, Y = np.meshgrid(np.arange(camera.width), self._lines)
        else:
            if not np.all(self._lines < camera.width):
                raise ValueError("Selected lines are outside the frame.")
            X, Y = np.meshgrid(self._lines, np.arange(camera.width))

        self.set_pixels(np.column_stack((X.ravel(), Y.ravel())), camera)

    def _select_random_lines(self, camera: Camera, dim: int = 0) -> np.ndarray:
        line_length = camera.width if dim == 0 else camera.height
        number_of_lines = self.number_of_defect_pixels(camera) // line_length
        selected_lines = self.rng.integers(line_length, size=(number_of_lines))
        return selected_lines


@dataclass
class RandomNoisePixelDefect(PixelDefect):
    """A pixel defect that introduces random noise to selected pixels.

    This is different from read noise, which is applied to all pixels and represents
    the electronic noise of the sensor. RandomNoisePixelDefect simulates pixels that
    are abnormally noisy (e.g., "noisy" or "unstable" pixels), and the noise is
    added only to those selected pixels.

    Examples
    --------
    >>> from cabaret import Observatory, Sources
    >>> pixel_defects = {"noise": {"type": "noise", "rate": 0.1, "noise_level": 5e3}}
    >>> observatory = Observatory(camera={"pixel_defects": pixel_defects})
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center
    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     exp_time=3, ra=ra, dec=dec, sources=sources, convert_all_to_adu=True
    ... )


    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 2, figsize=(7, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with random noise defects")
    >>> plt.subplots_adjust(wspace=0.1)
    >>> plt.show()
    """

    noise_level: float = 10.0  # Standard deviation for Gaussian noise
    """Standard deviation or scale for the noise."""

    distribution: Literal["normal", "poisson"] = "normal"
    """Distribution type for the noise ('normal' or 'poisson')."""

    def generate_noise(self, size: int) -> np.ndarray:
        """Generate noise for the defect pixels."""
        if self.distribution == "poisson":
            return self.noise_level * self.rng.poisson(size=size)
        elif self.distribution == "normal":
            return self.noise_level * self.rng.normal(size=size)
        else:
            raise ValueError("Unknown noise distribution.")

    def introduce_pixel_defect(self, image, camera, seed: int | None = None, **kwargs):
        if seed is not None:
            self._rng = numpy.random.default_rng(seed)

        if self._pixels is None:
            self.set_pixels(self._select_random_pixels(camera), camera)

        noise = self.generate_noise(self.pixels.shape[0])

        # Add noise only to the selected defect pixels
        image[self.pixels[:, 0], self.pixels[:, 1]] += noise
        image = np.clip(image, 0, camera.max_adu)

        return image


@dataclass
class QuantumEfficiencyMapPixelDefect(PixelDefect):
    """
    A pixel defect that simulates variations in quantum efficiency across the sensor.

    Examples
    --------
    >>> from cabaret import Observatory, Sources
    >>> pixel_defects = {
    ...     "qe": {"type": "quantum_efficiency_map", "quantum_efficiency_std": 0.5}
    ... }
    >>> observatory = Observatory(camera={"pixel_defects": pixel_defects})
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center
    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     exp_time=3, ra=ra, dec=dec, sources=sources, convert_all_to_adu=True
    ... )

    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 3, figsize=(10, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with QE defects")
    >>> _ = plot_image(
    ...     observatory.camera.pixel_defects["qe"].quantum_efficiency_map,
    ...     ax=axes[2], title="Quantum Efficiency Map"
    ... )
    >>> plt.subplots_adjust(wspace=0.3, hspace=0.3)
    >>> plt.show()
    """

    quantum_efficiency_std: float = 0.1
    """Standard deviation of the quantum efficiency variations."""

    quantum_efficiency_map: np.ndarray | None = None
    """Quantum efficiency map (2D array)."""

    def __post_init__(self):
        if self.quantum_efficiency_map is not None:
            self.quantum_efficiency_map = np.array(self.quantum_efficiency_map)

    def generate_quantum_efficiency_map(self, camera: Camera):
        self.quantum_efficiency_map = np.clip(
            self.rng.normal(
                loc=camera.average_quantum_efficiency,
                scale=self.quantum_efficiency_std,
                size=(camera.height, camera.width),
            ),
            0,
            1,
        )

    def introduce_pixel_defect(
        self, image: np.ndarray, camera: Camera, seed: int | None = None, **kwargs
    ):
        if seed is not None:
            self._rng = numpy.random.default_rng(seed)
        if self.quantum_efficiency_map is None:
            self.generate_quantum_efficiency_map(camera)

        image = image * self.quantum_efficiency_map / camera.average_quantum_efficiency
        image = np.clip(image, 0, camera.max_adu)
        return image


@dataclass
class ReadoutSmearPixelDefect(PixelDefect):
    """
    Simulates readout smear (frame transfer smear) caused by continued exposure during
    readout.

    Parameters
    ----------
    smear_fraction : float
        Fraction of the exposure time spent during readout (smear strength).
    dim : int
        Direction of readout: 0 for vertical (default), 1 for horizontal.
    readout_time : float
        Time taken to read out the entire frame (seconds).
    seed : int
        Random seed for reproducibility.

    Examples
    --------
    >>> from cabaret import Observatory, Sources
    >>> pixel_defects = {"smear": {"type": "readout_smear", "readout_time": 1}}
    >>> observatory = Observatory(camera={"pixel_defects": pixel_defects})
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.center
    >>> _, clean_image, image = observatory.generate_image_stack(
    ...     exp_time=5, ra=ra, dec=dec, sources=sources, convert_all_to_adu=True
    ... )


    To plot the images, you can use the `plot_image` function from `cabaret.plot`:

    >>> from cabaret.plot import plot_image
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(1, 2, figsize=(7, 5), sharex=True, sharey=True)
    >>> _ = plot_image(clean_image, ax=axes[0], title="Image without defects")
    >>> _ = plot_image(image, ax=axes[1], title="Image with readout smear")
    >>> plt.subplots_adjust(wspace=0.1)
    >>> plt.show()
    """

    readout_time: float = 0.1
    """Time taken to read out the entire frame (seconds)."""

    dim: int = 0
    """Direction of readout: 0 for vertical, 1 for horizontal."""

    def __post__init__(self):
        if self.dim not in (0, 1):
            raise ValueError("dim must be 0 (vertical) or 1 (horizontal)")

    def introduce_pixel_defect(self, image, camera, exp_time, **kwargs):
        # Fraction of the exposure time that each pixel is exposed
        # during the readout of a single row/column, assuming uniform readout speed
        smear_fraction_per_pixel_readout = (
            self.readout_time / exp_time / image.shape[self.dim]
        )

        # A fraction of each pixels values to those read out after it
        smear = smear_fraction_per_pixel_readout * np.cumsum(image, axis=self.dim)

        image = np.clip(image + smear, 0, camera.max_adu)
        return image



import logging
from datetime import UTC, datetime

import numpy as np
import numpy.random
from astropy import units as u
from astropy.coordinates import AltAz, Angle, EarthLocation, SkyCoord, get_sun
from astropy.time import Time
from astropy.wcs import WCS

from cabaret.camera import Camera
from cabaret.focuser import Focuser
from cabaret.queries import Filters, GaiaQuery
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope

logger = logging.getLogger("cabaret")


def moffat_profile(
    x: np.ndarray,
    y: np.ndarray,
    x0: float,
    y0: float,
    FWHM: float,
    beta: float = 2.5,
) -> np.ndarray:
    """
    Compute a normalized Moffat profile centered at (x0, y0).

    Parameters
    ----------
    x, y : np.ndarray
        Meshgrid arrays for the pixel coordinates.
    x0, y0 : float
        Center of the profile.
    FWHM : float
        Full width at half maximum of the profile.
    beta : float, optional
        Moffat beta parameter (default: 2.5).

    Returns
    -------
    np.ndarray
        Normalized Moffat profile evaluated on the grid.
    """
    # https://nbviewer.org/github/ysbach/AO_2017/blob/master/04_Ground_Based_Concept.ipynb#1.2.-Moffat
    # FWHM =  2 * R * (2**(1/beta) - 1)**0.5

    R = (FWHM / 2) * (1 / (2 ** (1 / beta) - 1) ** 0.5)
    A = (beta - 1) / (np.pi * R**2)

    r_squared = (x - x0) ** 2 + (y - y0) ** 2

    mp = A * (1 + (r_squared / R**2)) ** (-beta)

    mp_sum = np.sum(mp)

    return mp / mp_sum


def scintillation_noise(
    r: float,
    t: float,
    N_star: float,
    h: float = 2440,
    C: float = 1.56,
    airmass: float = 1.5,
) -> float:
    """
    Calculate the scintillation noise for a given set of parameters.

    Parameters
    ----------
    r : float
        Aperture radius in meters.
    t : float
        Exposure time in seconds.
    N_star : float
        Number of stars.
    h : float
        Altitude of the observatory in meters. Default is 2440 for Paranal Observatory.
    C : float
        Empirical coefficient. Default is 1.56, optimized for the
        20-cm NGTS telescopes at Paranal Observatory.
    airmass : float, optional
        Airmass value. Default is 1.5.

    Returns
    -------
    float
        The calculated scintillation noise.

    Reference
    ---------
    https://academic.oup.com/mnras/article/509/4/6111/6442285
    """

    return (
        np.sqrt(
            1e-5
            * C**2
            * pow(2 * r, -4 / 3)
            * t**-1
            * airmass**3
            * np.exp(-2 * h / 8000)
        )
        * N_star
        * t
    )


def generate_star_image(
    pos: np.ndarray,
    fluxes: list[float],
    FWHM: float,
    frame_size: tuple[int, int],
    rng: numpy.random.Generator,
    telescope_aperture: float,
    site_elevation: float,
    exp_time: float,
    airmass: float = 1.5,
    fwhm_multiplier: float = 5.0,
) -> np.ndarray:
    """
    Render stars onto an image using a fast, windowed approach.

    Parameters
    ----------
    pos : np.ndarray
        Pixel positions of stars (shape: 2, n_stars).
    fluxes : list
        list of fluxes for each star.
    FWHM : float
        Full width at half maximum for the Moffat profile.
    frame_size : tuple
        Size of the output image (width, height).
    rng : numpy.random.Generator
        Random number generator for Poisson noise.
    fwhm_multiplier : float, optional
        Multiplier to determine the rendering radius around each star (default: 5.0).

    Returns
    -------
    np.ndarray
        Image with rendered stars.
    """
    x = np.linspace(0, frame_size[0] - 1, frame_size[0])
    y = np.linspace(0, frame_size[1] - 1, frame_size[1])
    xx, yy = np.meshgrid(x, y)

    render_radius = FWHM * fwhm_multiplier  # render n * FWHM around the star

    image = np.zeros(frame_size).T
    for i, flux in enumerate(fluxes):
        x0 = pos[0][i]
        y0 = pos[1][i]
        if x0 < 0 or x0 >= frame_size[0] or y0 < 0 or y0 >= frame_size[1]:
            # print(f"Star {i} is outside the frame.")
            continue
        x_min, x_max = int(x0 - render_radius), int(x0 + render_radius)
        y_min, y_max = int(y0 - render_radius), int(y0 + render_radius)
        x_min, x_max = max(0, x_min), min(x_max, frame_size[0] - 1)
        y_min, y_max = max(0, y_min), min(y_max, frame_size[1] - 1)

        scint_noise = scintillation_noise(
            r=telescope_aperture / 2,
            t=exp_time,
            N_star=flux,
            h=site_elevation,
            C=1.56,
            airmass=airmass,
        )

        star_flux = rng.poisson(flux) + rng.normal(0, scint_noise)

        if star_flux < 0:
            star_flux = 0

        star = star_flux * moffat_profile(
            xx[y_min : y_max + 1, x_min : x_max + 1],
            yy[y_min : y_max + 1, x_min : x_max + 1],
            x0,
            y0,
            FWHM,
        )

        image[y_min : y_max + 1, x_min : x_max + 1] += star

    return image


def get_sources(
    center: SkyCoord,
    radius: Angle,
    dateobs: datetime,
    n_star_limit: int,
    filter_band: Filters | str,
    timeout: float | None,
    sources: Sources | None = None,
) -> Sources:
    """Get sources from Gaia or use provided sources."""
    if not isinstance(sources, Sources):
        logger.info("Querying Gaia for sources...")
        sources = GaiaQuery.get_sources(
            center=center,
            radius=radius,
            dateobs=dateobs,
            limit=n_star_limit,
            filter_band=filter_band,
            timeout=timeout,
        )
        logger.info(f"Found {len(sources)} sources (user set limit of {n_star_limit}).")
    return sources


def add_sun_sky_background(
    image: np.ndarray,
    site: Site,
    telescope: Telescope,
    camera: Camera,
    exp_time: float,
    dateobs: datetime,
    logger: logging.Logger,
) -> np.ndarray:
    """Add sky background and sunlight if location is specified."""

    # Sky brightness model due to sunlight scattering
    # calibrated for I+z band in Paranal
    SKY_MODEL_A = np.float64(4533508.655833181)
    SKY_MODEL_B = np.float64(0.3937301435229289)
    SKY_MODEL_C = np.float64(-0.7907223506021084)

    sun_altitude = _get_sun_altitude(site, dateobs, logger)

    if sun_altitude is None:
        return image

    sun_brightness = SKY_MODEL_A * SKY_MODEL_B ** (SKY_MODEL_C * sun_altitude)
    logger.debug(f"sun_brightness (ph/m2/arcsec2/s): {sun_brightness}")

    sun_electrons_per_sec = (
        sun_brightness
        * telescope.collecting_area
        * camera.average_quantum_efficiency
        * camera.plate_scale**2
    )
    logger.debug(f"sun_e (e-/s): {sun_electrons_per_sec}")

    sun_electrons = sun_electrons_per_sec * exp_time
    sun_background = np.random.poisson(
        np.ones((camera.height, camera.width), dtype=np.float64) * sun_electrons
    ).astype(np.float64)

    image += sun_background

    return image


def _get_sun_altitude(
    site: Site,
    dateobs: datetime,
    logger: logging.Logger,
) -> float | None:
    """
    Calculate the sun's altitude angle at the observation time.

    Parameters
    ----------
    site : Site
        Observatory site configuration.
    dateobs : datetime
        Observation date and time.
    logger : logging.Logger
        Logger instance for debug messages.

    Returns
    -------
    float or None
        Sun altitude in degrees, or None if it cannot be calculated.
    """
    if site.sun_altitude is not None:
        return site.sun_altitude

    if site.latitude is None or site.longitude is None:
        return None

    logger.debug("Calculating sunlight brightness based on sun's position")

    location = EarthLocation(
        lat=u.Quantity(site.latitude, "deg"),
        lon=u.Quantity(site.longitude, "deg"),
    )
    obs_time = Time(dateobs, scale="utc")
    sun = get_sun(obs_time)
    altaz_frame = AltAz(obstime=obs_time, location=location)
    sun_altaz = sun.transform_to(altaz_frame)
    sun_altitude: float = sun_altaz.alt.degree  # type: ignore

    logger.debug(f"Sun altitude: {sun_altitude:.5f} deg at {dateobs} UTC")

    return sun_altitude


def add_stars(
    image: np.ndarray,
    sources: Sources,
    camera: Camera,
    focuser: Focuser,
    telescope: Telescope,
    site: Site,
    exp_time: float,
    rng: numpy.random.Generator,
    ra: float | None,
    dec: float | None,
    wcs: WCS | None,
    dateobs: datetime | None,
    airmass: float | None,
    fwhm_multiplier: float = 5.0,
) -> np.ndarray:
    """Add stars to the image using the Moffat profile and sky background."""
    if len(sources) > 0:
        fluxes = (
            sources.fluxes
            * camera.average_quantum_efficiency
            * telescope.collecting_area
            * exp_time
        )  # [electrons]

        if ra is None or dec is None:
            ra, dec = sources.center

        if wcs is None:
            wcs = camera.get_wcs(SkyCoord(ra=ra, dec=dec, unit="deg"))
        gaias_pixel = sources.to_pixel(wcs)
        on_camera_mask = camera.on_camera_mask(gaias_pixel)
        logger.debug(
            f"Number of stars on camera: {on_camera_mask.sum()}"
            f" out of {len(sources)} total stars "
            f"({on_camera_mask.sum() / len(sources) * 100})%."
        )

        if airmass is not None:
            logger.debug(f"Using provided airmass: {airmass:.3f}.")
        elif site.latitude and site.longitude:
            location = EarthLocation(
                lat=u.Quantity(site.latitude, "deg"),
                lon=u.Quantity(site.longitude, "deg"),
            )
            altitude_azimuth_frame = AltAz(obstime=Time(dateobs), location=location)
            center_coord = SkyCoord(ra=ra, dec=dec, unit="deg")
            center_altaz = center_coord.transform_to(altitude_azimuth_frame)
            zenith_angle = 90.0 - center_altaz.alt.degree  # degrees
            airmass = 1.0 / np.cos(np.radians(zenith_angle))

            if airmass < 1.0:
                logger.warning(
                    f"Calculated airmass {airmass:.3f} is less than 1.0."
                    " Setting airmass to 1.0."
                )
                airmass = 1.0  # minimum airmass is 1.0
        else:
            airmass = 1.5  # default value

        stars = generate_star_image(
            gaias_pixel[:, on_camera_mask],
            fluxes[on_camera_mask],
            focuser.seeing_multiplier * site.seeing / camera.plate_scale,
            (camera.width, camera.height),
            rng=rng,
            telescope_aperture=telescope.diameter,
            site_elevation=site.elevation if site.elevation is not None else 0,
            exp_time=exp_time,
            airmass=airmass,
            fwhm_multiplier=fwhm_multiplier,
        ).astype(np.float64)

        sky_background = (
            site.sky_background * telescope.collecting_area * camera.plate_scale**2
        )  # [e-/s]

        image = image + rng.poisson(
            np.ones((camera.height, camera.width)).astype(np.float64)
            * sky_background
            * exp_time
        ).astype(np.float64)

        image += stars
    return image


def add_stars_and_sky(
    base: np.ndarray,
    ra: float | None,
    dec: float | None,
    exp_time: float,
    dateobs: datetime | None,
    light: int,
    camera: Camera,
    focuser: Focuser,
    telescope: Telescope,
    site: Site,
    filter_band: Filters | str,
    airmass: float,
    n_star_limit: int,
    rng: numpy.random.Generator,
    timeout: float | None,
    sources: Sources | None,
    wcs: WCS | None,
    fwhm_multiplier: float = 5.0,
) -> np.ndarray:
    """Add stars and sky background to the base image."""
    if light == 1:
        if ra is None and dec is None and not isinstance(sources, Sources):
            raise ValueError("Either ra/dec or sources must be provided for light.")
        if ra is None or dec is None and isinstance(sources, Sources):
            ra, dec = sources.center  # type: ignore
        center = SkyCoord(ra=ra, dec=dec, unit="deg")

        assert camera.plate_scale is not None, "Camera plate scale must be set."
        radius = camera.get_fov_radius()

        if dateobs is None:
            dateobs = datetime.now(UTC)
        sources = get_sources(
            center=center,
            radius=radius,
            dateobs=dateobs,
            n_star_limit=n_star_limit,
            filter_band=filter_band,
            timeout=timeout,
            sources=sources,
        )
        image = base
        image = add_sun_sky_background(
            image, site, telescope, camera, exp_time, dateobs, logger
        )
        image = add_stars(
            image=image,
            sources=sources,
            camera=camera,
            focuser=focuser,
            telescope=telescope,
            site=site,
            exp_time=exp_time,
            rng=rng,
            ra=ra,
            dec=dec,
            wcs=wcs,
            dateobs=dateobs,
            airmass=airmass,
            fwhm_multiplier=fwhm_multiplier,
        )
    else:
        image = base
    return image


def generate_image(
    ra: float,
    dec: float,
    exp_time: float,
    dateobs: datetime | None = None,
    light: int = 1,
    camera: Camera = Camera(),
    focuser: Focuser = Focuser(),
    telescope: Telescope = Telescope(),
    site: Site = Site(),
    filter_band: Filters | str = Filters.G,
    airmass: float = 1.5,
    n_star_limit: int = 2000,
    rng: numpy.random.Generator = numpy.random.default_rng(),
    seed: int | None = None,
    timeout: float | None = None,
    sources: Sources | None = None,
    wcs: WCS | None = None,
    fwhm_multiplier: float = 5.0,
) -> np.ndarray:
    """
    Generate a simulated astronomical image.

    Parameters
    ----------
    ra : float
        Right ascension of the image center (degrees).
    dec : float
        Declination of the image center (degrees).
    exp_time : float
        Exposure time in seconds.
    dateobs : datetime, optional
        Observation date and time (default: now, UTC).
    light : int, optional
        If 1, simulate light exposure; if 0, simulate dark exposure.
    camera : Camera, optional
        Camera configuration.
    focuser : Focuser, optional
        Focuser configuration.
    telescope : Telescope, optional
        Telescope configuration.
    site : Site, optional
        Observatory site configuration.
    filter_band : Filters or str, optional
        The filter to use for the flux column. Default is "G".
    airmass : float, optional
        Airmass at the image center (default: 1.5).
    n_star_limit : int, optional
        Maximum number of stars to simulate.
    rng : numpy.random.Generator, optional
        Random number generator.
    seed : int or None, optional
        Seed for the random number generator.
    timeout : float or None, optional
        Timeout for Gaia query.
    sources : Sources or None, optional
        Precomputed sources to use instead of querying Gaia.
    wcs : WCS or None, optional
        World Coordinate System information for the image.
    fwhm_multiplier : float, optional
        Multiplier to determine the rendering radius around each star (default: 5.0).

    Returns
    -------
    np.ndarray
        Simulated image as a 2D numpy array (uint16).
    """
    if seed is not None:
        rng = numpy.random.default_rng(seed)

    if camera.plate_scale is None:
        camera.set_plate_scale_from_focal_length(telescope.focal_length)

    base = camera.make_base_image(exp_time=exp_time, rng=rng)

    if light == 1:
        image = add_stars_and_sky(
            base=base,
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=camera,
            focuser=focuser,
            telescope=telescope,
            site=site,
            filter_band=filter_band,
            airmass=airmass,
            n_star_limit=n_star_limit,
            rng=rng,
            timeout=timeout,
            sources=sources,
            wcs=wcs,
            fwhm_multiplier=fwhm_multiplier,
        )
    else:
        image = base

    image = camera.apply_pixel_defects(image, exp_time)

    image = camera.bin_image(image)

    image = camera.to_adu_image(image)

    return image


def generate_image_stack(
    ra: float,
    dec: float,
    exp_time: float,
    dateobs: datetime | None = None,
    light: int = 1,
    camera: Camera = Camera(),
    focuser: Focuser = Focuser(),
    telescope: Telescope = Telescope(),
    site: Site = Site(),
    filter_band: Filters | str = Filters.G,
    airmass: float = 1.5,
    n_star_limit: int = 2000,
    rng: numpy.random.Generator = numpy.random.default_rng(),
    seed: int | None = None,
    timeout: float | None = None,
    sources: Sources | None = None,
    convert_all_to_adu: bool = False,
    wcs: WCS | None = None,
    fwhm_multiplier: float = 5.0,
) -> np.ndarray:
    """
    Generate a stack of images from different stages in the image simulation pipeline.

    Parameters
    ----------
    ra : float
        Right ascension of the image center (degrees).
    dec : float
        Declination of the image center (degrees).
    exp_time : float
        Exposure time in seconds.
    dateobs : datetime, optional
        Observation date and time (default: now, UTC).
    light : int, optional
        If 1, simulate light exposure; if 0, simulate dark exposure.
    camera : Camera, optional
        Camera configuration.
    focuser : Focuser, optional
        Focuser configuration.
    telescope : Telescope, optional
        Telescope configuration.
    site : Site, optional
        Observatory site configuration.
    filter_band : Filters or str, optional
        The filter to use for the flux column. Default is "G".
    airmass : float, optional
        Airmass at the image center (default: 1.5).
    n_star_limit : int, optional
        Maximum number of stars to simulate.
    rng : numpy.random.Generator, optional
        Random number generator.
    seed : int or None, optional
        Seed for the random number generator.
    timeout : float or None, optional
        Timeout for Gaia query.
    sources : Sources or None, optional
        Precomputed sources to use instead of querying Gaia.
    convert_all_to_adu : bool, optional
        Whether to convert all images to ADU. Default is False.
    wcs : WCS or None, optional
        World Coordinate System information for the image.
    fwhm_multiplier : float, optional
        Multiplier to determine the rendering radius around each star (default: 5.0).

    Returns
    -------
    np.ndarray
        Simulated image stack as a 3D numpy array (uint16, shape (3, height, width)).
        The first slice is the base image, the second is the astronomical image,
        and the third is the ADU image with pixel defects applied.


    """
    if seed is not None:
        rng = numpy.random.default_rng(seed)

    if camera.plate_scale is None:
        camera.set_plate_scale_from_focal_length(telescope.focal_length)

    base = camera.make_base_image(exp_time=exp_time, rng=rng)

    if light == 1:
        image = add_stars_and_sky(
            base=np.zeros_like(base),
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=camera,
            focuser=focuser,
            telescope=telescope,
            site=site,
            filter_band=filter_band,
            airmass=airmass,
            n_star_limit=n_star_limit,
            rng=rng,
            timeout=timeout,
            sources=sources,
            wcs=wcs,
            fwhm_multiplier=fwhm_multiplier,
        )
    else:
        image = base

    adu_image = camera.apply_pixel_defects(image.copy(), exp_time)

    adu_image = camera.to_adu_image(adu_image)

    if convert_all_to_adu:
        base = camera.to_adu_image(base)
        image = camera.to_adu_image(image)

    return np.stack([base, image, adu_image])


if __name__ == "__main__":
    import importlib.util

    camera = Camera(width=2000, height=2000)
    telescope = Telescope()
    site = Site(seeing=1.3, sky_background=350)
    exp_time = 0.1  # [s]

    logger.info("Generating image...")

    # example usage
    image = generate_image(
        ra=323.36152,
        dec=-0.82325,
        exp_time=exp_time,
        camera=camera,
        telescope=telescope,
        site=site,
    )

    science = image  # - camera.dark_current / camera.gain * exp_time - camera.bias

    if importlib.util.find_spec("matplotlib") is not None:
        import matplotlib.pyplot as plt

        from cabaret.plot import plot_image

        print("Plotting image...")
        med = np.median(science)
        std = np.std(science)
        print(med, std)

        fig, ax = plt.subplots()
        plot_image(image, ax=ax, title="Simulated Image")
        plt.show()



from dataclasses import dataclass


@dataclass
class Focuser:
    """
    A simple focuser model for simulating the effect of defocus on image quality.
    """

    position: float = 10_000
    """The current focus position of the focuser."""

    best_position: float = 10_000
    """The optimal focus position for best image quality."""

    scale: float = 100
    """The scale factor that determines how quickly defocus degrades seeing."""

    max_seeing_multiplier: float = 5.0
    """The maximum factor by which seeing can be increased due to defocus."""

    @property
    def seeing_multiplier(
        self,
    ) -> float:
        """Factor by which the seeing is increased due to defocus."""
        offset = abs(self.position - self.best_position)
        return min(1 + offset / self.scale, self.max_seeing_multiplier)



import numpy as np
import pytest

from cabaret.camera import Camera, ConstantPixelDefect


def test_camera_with_pixel_defects():
    defects = {"hot": {"rate": 0.1, "seed": 42}, "cold": {"rate": 0.05, "seed": 43}}
    camera = Camera(pixel_defects=defects)
    assert len(camera.pixel_defects) == 2
    assert all(
        isinstance(defect, ConstantPixelDefect)
        for defect in camera.pixel_defects.values()
    )


def test_constant_pixel_defect_set_pixels():
    camera = Camera(width=10, height=10)
    defect = ConstantPixelDefect(value=255, seed=42)

    # Valid case
    valid_pixels = np.array([[0, 0], [1, 1], [2, 2]])
    defect.set_pixels(valid_pixels, camera)
    assert np.array_equal(defect.pixels, valid_pixels)

    # Invalid case: Out of bounds
    with pytest.raises(ValueError, match="defect pixels are outside the frame."):
        out_of_bounds_pixels = np.array([[10, 10]])
        defect.set_pixels(out_of_bounds_pixels, camera)


if __name__ == "__main__":
    pytest.main()



from datetime import datetime
from pathlib import Path
from typing import Any

import numpy as np
from astropy.io import fits
from astropy.wcs import WCS


class FITSManager:
    """A class to manage FITS file operations, including creating headers with
    observatory metadata.

    Examples
    --------
    >>> from cabaret.fits_manager import FITSManager
    >>> from cabaret import Observatory, Sources
    >>> observatory = Observatory(name="My Observatory")
    >>> sources = Sources.get_test_sources()
    >>> ra, dec = sources.ra.deg.mean(), sources.dec.deg.mean()
    >>> image = observatory.generate_image(
    ...     ra=ra, dec=dec, exp_time=30, sources=sources, seed=0
    ... )
    >>> hdu_list = FITSManager.get_fits_from_array(
    ...     image=image, observatory=observatory, exp_time=30, ra=ra, dec=dec
    ... )
    >>> hdu_list[0].header['DEC']

    """

    @staticmethod
    def save(
        observatory,
        file_path: str | Path,
        hdu_list: fits.HDUList | None = None,
        image: np.ndarray | None = None,
        user_header: dict[str, Any] | fits.Header | None = None,
        exp_time: float | None = None,
        ra: float | None = None,
        dec: float | None = None,
        dateobs: datetime | None = None,
        wcs: WCS | None = None,
        overwrite: bool = True,
    ) -> fits.HDUList:
        """
        Save a numpy array image to a FITS file using observatory metadata.

        Parameters
        ----------
        observatory : Observatory
            The observatory instance for header metadata.
        file_path : str
            The path to the FITS file to write.
        hdu_list : fits.HDUList, optional
            An existing HDUList to save. If provided, image and header parameters
            are ignored.
        image : numpy.ndarray
            The image data to save.
        user_header : dict or fits.Header, optional
            Additional header keywords to add.
        exp_time : float, optional
            Exposure time in seconds. Passed to the header.
        ra : float, optional
            Right ascension of the image center in degrees. Passed to the header.
        dec : float, optional
            Declination of the image center in degrees. Passed to the header.
        dateobs : datetime, optional
            The observation date and time. Passed to the header.
        wcs : WCS, optional
            World Coordinate System information for the image. Passed to the header.
        overwrite : bool, optional
            Whether to overwrite existing file (default: True).
        """
        if hdu_list is None:
            hdu_list = FITSManager.to_hdu_list(
                image=image,
                observatory=observatory,
                exp_time=exp_time,
                ra=ra,
                dec=dec,
                dateobs=dateobs,
                wcs=wcs,
                user_header=user_header,
            )
        elif not isinstance(hdu_list, fits.HDUList):
            raise TypeError("hdu_list must be an instance of fits.HDUList.")

        hdu_list.writeto(file_path, overwrite=overwrite)

        return hdu_list

    @staticmethod
    def to_hdu_list(
        observatory,
        image,
        exp_time: float | None = None,
        ra: float | None = None,
        dec: float | None = None,
        dateobs: datetime | None = None,
        wcs: WCS | None = None,
        user_header: dict[str, Any] | fits.Header | None = None,
    ) -> fits.HDUList:
        """
        Create a FITS HDUList from a numpy array, optionally with a header.

        Parameters
        ----------
        observatory : Observatory
            The observatory instance for header metadata.
        image : numpy.ndarray
            The image data to convert to FITS.
        exp_time : float, optional
            Exposure time in seconds.
        ra : float, optional
            Right ascension of the image center in degrees.
        dec : float, optional
            Declination of the image center in degrees.
        dateobs : datetime, optional
            The observation date and time.
        wcs : WCS, optional
            World Coordinate System information for the image.
        user_header : dict or fits.Header, optional
            Additional header keywords to add.

        Returns
        -------
        fits.HDUList
            The created FITS HDUList.
        """

        header = FITSManager.get_header_from_observatory(
            observatory,
            user_header=user_header,
        )
        FITSManager.add_image_info_to_header(
            header, exp_time=exp_time, ra=ra, dec=dec, dateobs=dateobs, wcs=wcs
        )
        hdu = fits.PrimaryHDU(data=image, header=header)
        hdul = fits.HDUList([hdu])
        return hdul

    @staticmethod
    def add_image_info_to_header(
        header: fits.Header,
        exp_time: float | None = None,
        ra: float | None = None,
        dec: float | None = None,
        dateobs: datetime | None = None,
        wcs: WCS | None = None,
    ):
        """Add image-specific info to a FITS header."""
        if exp_time is not None:
            header["EXPTIME"] = (float(exp_time), "Exposure time in seconds")
        if ra is not None:
            header["RA"] = (float(ra), "Right ascension of image center [deg]")
        if dec is not None:
            header["DEC"] = (float(dec), "Declination of image center [deg]")
        if dateobs is None:
            dateobs = datetime.now()
        if wcs is not None:
            wcs_header = wcs.to_header()
            for card in wcs_header.cards:
                header[card.keyword] = (card.value, card.comment)
        header["DATE-OBS"] = (dateobs.isoformat(), "UTC datetime start of exposure")

    @staticmethod
    def get_header_from_observatory(
        observatory,
        user_header: dict[str, Any] | fits.Header | None = None,
    ) -> "fits.Header":
        """
        Create a FITS header populated with metadata from an Observatory instance.

        Parameters
        ----------
        observatory : Observatory
            The observatory instance to extract metadata from.
        extra : dict, optional
            Additional header keywords to add.

        Returns
        -------
        fits.Header
            The populated FITS header.
        """
        header = fits.Header()
        header["OBSNAME"] = (observatory.name, "Observatory name")

        # Camera-related fields
        camera = observatory.camera
        header["INSTRUME"] = (getattr(camera, "name", "Camera"), "Instrument used")
        header["XBINNING"] = (
            getattr(camera, "bin_x", 1),
            "Binning level along the X-axis",
        )
        header["YBINNING"] = (
            getattr(camera, "bin_y", 1),
            "Binning level along the Y-axis",
        )
        header["XPIXSZ"] = (
            getattr(camera, "pitch", None),
            "Pixel Width in microns (after binning)",
        )
        header["YPIXSZ"] = (
            getattr(camera, "pitch", None),
            "Pixel Height in microns (after binning)",
        )
        header["CAM-DNAM"] = (
            getattr(camera, "name", "cabaret_camera"),
            "Short name of Camera driver",
        )

        header["FOCUSER"] = (
            getattr(observatory.focuser, "name", "cabaret_focuser"),
            "Focuser name",
        )
        header["FOCUSPOS"] = (observatory.focuser.position, "Focuser position")
        header["TELESCOP"] = (
            getattr(observatory.telescope, "name", "cabaret_telescope"),
            "Telescope name",
        )
        telescope = observatory.telescope
        header["FOCALLEN"] = (
            telescope.focal_length,
            "[m] Focal length of telescope",
        )
        header["APTDIA"] = (
            telescope.diameter,
            "[m] Aperture diameter of telescope",
        )
        header["APTAREA"] = (
            telescope.collecting_area,
            "[m^2] Aperture area of telescope",
        )
        header["SITE"] = (
            getattr(observatory.site, "name", "cabaret_site"),
            "Site name",
        )
        # Add more fields as needed from the observatory/camera/telescope/site

        if user_header:
            # Accept both dict and fits.Header
            if isinstance(user_header, dict):
                for k, v in user_header.items():
                    header[k] = v
            elif isinstance(user_header, fits.Header):
                for card in user_header.cards:
                    header[card.keyword] = (card.value, card.comment)
            else:
                raise TypeError("user_header must be a dict or fits.Header")

        try:
            from importlib.metadata import version

            header["CABARET"] = (version("cabaret"), "Version of Cabaret")
        except Exception:
            header["CABARET"] = ("unknown", "Version of Cabaret")
        return header



from datetime import datetime

import numpy as np
import pytest

from cabaret.camera import Camera
from cabaret.fits_manager import FITSManager
from cabaret.observatory import Observatory


@pytest.fixture
def observatory():
    return Observatory(
        name="TestObs",
        camera=Camera(width=10, height=10),
    )


def test_get_header_from_observatory(observatory):
    header = FITSManager.get_header_from_observatory(observatory)
    assert header["OBSNAME"] == "TestObs"
    assert "INSTRUME" in header
    assert "FOCUSER" in header
    assert "TELESCOP" in header
    assert "SITE" in header
    assert "CABARET" in header


def test_add_image_info_to_header():
    from astropy.io import fits

    header = fits.Header()
    dateobs = datetime(2024, 1, 1, 12, 0, 0)
    FITSManager.add_image_info_to_header(
        header, exp_time=10, ra=123.4, dec=56.7, dateobs=dateobs
    )
    assert header["EXPTIME"] == 10.0
    assert header["RA"] == 123.4
    assert header["DEC"] == 56.7
    assert header["DATE-OBS"].startswith("2024-01-01T12:00:00")


def test_to_hdu_list_and_save(tmp_path, observatory):
    image = np.ones((10, 10), dtype=np.uint16)
    ra, dec = 10.0, 20.0
    hdul = FITSManager.to_hdu_list(
        image=image,
        observatory=observatory,
        exp_time=30,
        ra=ra,
        dec=dec,
        dateobs=datetime(2024, 1, 1, 12, 0, 0),
        user_header={"TESTKEY": ("testval", "test comment")},
    )
    assert hdul[0].header["EXPTIME"] == 30.0
    assert hdul[0].header["RA"] == 10.0
    assert hdul[0].header["DEC"] == 20.0
    assert hdul[0].header["TESTKEY"] == "testval"

    # Test saving to file
    file_path = tmp_path / "test_image.fits"
    FITSManager.save(
        image=image,
        file_path=str(file_path),
        observatory=observatory,
        exp_time=30,
        ra=ra,
        dec=dec,
        dateobs=datetime(2024, 1, 1, 12, 0, 0),
        user_header={"TESTKEY": ("testval", "test comment")},
        overwrite=True,
    )
    from astropy.io import fits as afits

    with afits.open(file_path) as hdul2:
        assert hdul2[0].header["EXPTIME"] == 30.0
        assert hdul2[0].header["RA"] == 10.0
        assert hdul2[0].header["DEC"] == 20.0
        assert hdul2[0].header["TESTKEY"] == "testval"
        np.testing.assert_array_equal(hdul2[0].data, image)



import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS

from cabaret.sources import Sources

COORDS = np.array([[10.64, 41.26], [10.68, 41.22]])
FLUXES = np.array([169435.6, 52203.9])


@pytest.fixture
def example_sources():
    return Sources(SkyCoord(COORDS, unit="deg"), FLUXES)


def test_sources_init_and_len(example_sources):
    sources = example_sources
    assert len(sources) == 2
    assert np.allclose(sources.fluxes, FLUXES)
    assert np.allclose(sources.coords.ra.deg, COORDS[:, 0])
    assert np.allclose(sources.coords.dec.deg, COORDS[:, 1])


def test_sources_ra_dec_properties(example_sources):
    sources = example_sources
    assert np.allclose(sources.ra.deg, COORDS[:, 0])
    assert np.allclose(sources.dec.deg, COORDS[:, 1])


def test_sources_to_pixel(example_sources):
    sources = example_sources
    wcs = WCS(naxis=2)
    wcs.wcs.crval = [10.65, 41.25]
    wcs.wcs.crpix = [100, 100]
    wcs.wcs.cdelt = [-0.0002777778, 0.0002777778]
    wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    pixels = sources.to_pixel(wcs)
    assert pixels.shape == (2, 2)


def test_sources_invalid_init():
    # coords is not of type SkyCoord
    with pytest.raises(ValueError):
        Sources(COORDS, FLUXES)

    # Length mismatch
    with pytest.raises(ValueError):
        Sources(SkyCoord(COORDS, unit="deg"), FLUXES[:-1])


def test_sources_from_array():
    sources = Sources.from_arrays(COORDS[:, 0], COORDS[:, 1], FLUXES)
    assert np.allclose(sources.ra.deg, COORDS[:, 0])
    assert np.allclose(sources.dec.deg, COORDS[:, 1])
    assert np.allclose(sources.fluxes, FLUXES)


def test_center():
    coords = SkyCoord(ra=[179.999, 180.001], dec=[10.68, 10.689], unit="deg")
    sources = Sources(coords, fluxes=np.array([169_435.6, 92_203.9]))

    ra_center, dec_center = sources.center
    assert np.isclose(ra_center, 180.0)
    assert np.isclose(dec_center, 10.6845)


def test_center_with_wrap():
    coords = SkyCoord(ra=[359.999, 0.001], dec=[10.68, 10.689], unit="deg")
    sources = Sources(coords, fluxes=np.array([169_435.6, 92_203.9]))

    ra_center, _ = sources.center
    assert np.isclose(ra_center, 0.0)
